#include <sys/types.h>
#include <sys/mman.h>
#include <sys/stat.h>

#include <limits.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <getopt.h>
#include <locale.h>

#include <Python.h>

#include <tremor/ivorbisfile.h>

/* We should probably make that an optional argument to the constructor */
#define BUFFERSIZE 8192

#define ERR_CLEARTEXT(code) (code==OV_EREAD?"Read error":\
		code==OV_ENOTVORBIS?"Not a vorbis file":\
		code==OV_EVERSION?"Unsupported vorbis file":\
		code==OV_EBADHEADER?"Bad vorbis header":\
		code==OV_EFAULT?"Disk and/or RAM on fire":\
		"Unknown Error")

typedef struct {
	PyObject_HEAD
	char *fName;
	OggVorbis_File *vorbisDec;
	unsigned char *buffer;
	int bufferSize;
} vorbisobject;

staticforward PyTypeObject VorbisFileType;

static PyObject*
makeVorbisStruct(char *fName, int bufferSize)
{
	vorbisobject *self;
	FILE *input;
	int res;

	if (!(input = fopen(fName, "r"))) {
		PyErr_SetFromErrno(PyExc_IOError);
		return NULL;
	}
	if (!(self = PyObject_New(vorbisobject, &VorbisFileType)) ||
			!(self->fName = strdup(fName)) ||
			!(self->buffer = malloc(bufferSize)) ||
			!(self->vorbisDec = malloc(sizeof(OggVorbis_File)))) {
				return PyErr_NoMemory(); 
				/* This will probably leak memory, but at that point, we'll
				 * want to terminate anyway. */
	}
	self->bufferSize = bufferSize;
	if ((res=ov_open(input, self->vorbisDec, NULL, 0))) {
		Py_DECREF(self);
		PyErr_Format(PyExc_IOError, ERR_CLEARTEXT(res));
		return NULL;
	}
	return (PyObject*)self;
}

static PyObject*
vorbisfile_new(PyObject *self, PyObject *args)
{
	char *fName;
	int buffersize=BUFFERSIZE;

	if (!PyArg_ParseTuple(args, "s|i", &fName, &buffersize)) {
		return NULL;
	}
	return makeVorbisStruct(fName, buffersize);
}

static PyObject*
vorbisfile_read(PyObject *self_o, PyObject *args)
{
	int bytesRequested, bytesRead, streamIndex;
	vorbisobject *self=(vorbisobject*)self_o;

	if (!PyArg_ParseTuple(args, "i", &bytesRequested)) {
		return NULL;
	}
	if (bytesRequested>self->bufferSize) {
		bytesRequested = self->bufferSize;
	}
	bytesRead = ov_read(self->vorbisDec, self->buffer, 
		bytesRequested, &streamIndex);
	if (bytesRead<0) {
		return PyErr_Format(PyExc_IOError, "Read error on vorbis stream");
	} else if (bytesRead==0) {
		return Py_BuildValue("Oi", Py_None, -1L);
	} else {
		return Py_BuildValue("Ni", PyString_FromStringAndSize(
			self->buffer, bytesRead), -1L);
	}
	/* Can't happen (if it does, you'll see funny exceptions...) */
	return NULL;
}


static PyObject*
vorbisfile_getTotalTime(PyObject *self_o, PyObject *args) 
{
	int bitstreamInd=-1;
	ogg_int64_t lengthMs;
	vorbisobject *self=(vorbisobject*)self_o;

	if (!PyArg_ParseTuple(args, "|i", &bitstreamInd)) {
		return NULL;
	}
	lengthMs = ov_time_total(self->vorbisDec, bitstreamInd);
	if (lengthMs<0) {
		return PyErr_Format(PyExc_ValueError, "Cannot compute length"
			" for bitstream %d", bitstreamInd);
	} else {
		/* 31 bits are plenty enough if you count in seconds -- it's about
		 * 68 years, and your computer will have problems other than failing
		 * totalTimes on that timescale.
		 */
		int lengthSec=lengthMs/1000;
		return PyInt_FromLong(lengthSec);
	}
}


static PyObject*
vorbisfile_tellTime(PyObject *self_o, PyObject *args) 
{
	ogg_int64_t timeMs;
	vorbisobject *self=(vorbisobject*)self_o;

	if (!PyArg_ParseTuple(args, "")) {
		return NULL;
	}
	timeMs = ov_time_tell(self->vorbisDec);
	if (timeMs<0) {
		return PyErr_Format(PyExc_ValueError, "Cannot compute current time");
	} else {
		int timeSec=timeMs/1000;
		return PyInt_FromLong(timeSec);
	}
}


static PyObject*
vorbisfile_getPars(PyObject *self_o, PyObject *args) 
{
	vorbis_info *info;
	vorbisobject *self=(vorbisobject*)self_o;

	if (!PyArg_ParseTuple(args, "")) {
		return NULL;
	}
	if (!(info=ov_info(self->vorbisDec, -1))) {
		return PyErr_Format(PyExc_ValueError, "Invalid vorbis stream");
	}
	return Py_BuildValue("ii", info->channels, info->rate);
}	


static PyObject*
vorbisfile_seek(PyObject *self_o, PyObject *args)
{
	int targetSecs;
	ogg_int64_t targetMillisecs;
	vorbisobject *self=(vorbisobject*)self_o;

	if (!PyArg_ParseTuple(args, "i", &targetSecs)) {
		return NULL;
	}
	targetMillisecs = (ogg_int64_t)targetSecs*1000;
	if (ov_time_seek(self->vorbisDec, targetMillisecs)) {
		/* FIXME: can return OV_EINVAL, OV_ENOSEEK */
		return PyErr_Format(PyExc_IOError, "Could not seek.");
	}
	Py_INCREF(Py_None);
	return Py_None;
}


/* Method table for the VorbisFile class, for getattr */
static struct PyMethodDef vorbisfile_methods[]={
{"read", (PyCFunction)vorbisfile_read, METH_VARARGS, 
	"returns a tuple with the next data and the current stream index."},
{"getTotalTime", (PyCFunction)vorbisfile_getTotalTime, METH_VARARGS, 
	"returns the length of the stream in seconds.  Pass an optional\n"
	"bitstream index to obtain the length of a logical bitstream."},
{"tellTime", (PyCFunction)vorbisfile_tellTime, METH_VARARGS, 
	"returns the current decoding position in seconds."},
{"getPars", (PyCFunction)vorbisfile_getPars, METH_VARARGS, 
	"returns the current number of channels and samplerate."},
{"seek", (PyCFunction)vorbisfile_seek, METH_VARARGS, 
	"seeks to argument seconds in the current stream."},
{NULL,NULL, 0, NULL},
};

/* getattr is just a dispatcher here */
static PyObject *vorbisfile_getattr(PyObject *self, char *name)
{	
	return Py_FindMethod(vorbisfile_methods, self, name);
}

/* the destructor needs to free all C allocated memory */
static void vorbisfile_destroy(PyObject *self_o)
{ 
	vorbisobject *self=(vorbisobject*)self_o;
	ov_clear(self->vorbisDec);
	free(self->fName);
	free(self->buffer);
	free(self->vorbisDec);
	PyObject_Del(self);
}

static PyTypeObject VorbisFileType = {
	PyObject_HEAD_INIT(&PyType_Type)
	0,  /* never mind */
	"VorbisFile", /* Name */
	sizeof(VorbisFileType),
	0,  /* never mind */
	(destructor)vorbisfile_destroy,
	(printfunc)0,
	(getattrfunc)vorbisfile_getattr,
	(setattrfunc)0,
	(cmpfunc)0,
	(reprfunc)0,
	0, /* as number */
	0, /* as sequence */
	0, /* as mapping */
};

/*
-------------------- Python interface, Module logic -------------------
*/

static PyMethodDef tremorMethods[] = {
	{"VorbisFile", vorbisfile_new, METH_VARARGS, 
		"creates a new VorbisFile instance."},
	{NULL, NULL, 0, NULL},
};

void initpytremor(void)
{
	Py_InitModule("pytremor", tremorMethods);
}
